@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2023
-;; Version: 0.7
+;; Version: 0.8
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.4.0"))
 ;; Homepage: https://github.com/minad/jinx
 ;; Keywords: convenience, wp
@@ -52,7 +52,9 @@
 ;;; Code:
 
 (require 'compat)
-(eval-when-compile (require 'cl-lib))
+(eval-when-compile
+  (require 'cl-lib)
+  (require 'subr-x))
 
 ;;;; Customization
 
@@ -68,24 +70,6 @@
   "Idle timer delay."
   :type 'float)
 
-(defface jinx-misspelled
-  '((((class color) (min-colors 88) (background dark)
-      (supports :underline (:style wave)))
-     :underline (:style wave :color "#d2b580"))
-    (((class color) (min-colors 88) (background light)
-      (supports :underline (:style wave)))
-     :underline (:style wave :color "#5f4400"))
-    (t :underline t :inherit error))
-  "Face used for misspelled words.")
-
-(defface jinx-highlight
-  '((t :inherit isearch))
-  "Face used to highlight current misspelling during correction.")
-
-(defface jinx-accept
-  '((t :inherit font-lock-negation-char-face))
-  "Face used for the accept action during correction.")
-
 (defcustom jinx-languages
   (replace-regexp-in-string
    "\\..*\\'" ""
@@ -179,6 +163,39 @@ checking."
 ;;;###autoload
 (put 'jinx-local-words 'safe-local-variable #'stringp)
 
+;;;; Faces
+
+(defgroup jinx-faces nil
+  "Faces used by Jinx."
+  :group 'jinx
+  :group 'faces)
+
+(defface jinx-misspelled
+  '((((class color) (min-colors 88) (background dark)
+      (supports :underline (:style wave)))
+     :underline (:style wave :color "#d2b580"))
+    (((class color) (min-colors 88) (background light)
+      (supports :underline (:style wave)))
+     :underline (:style wave :color "#5f4400"))
+    (t :underline t :inherit error))
+  "Face used for misspelled words.")
+
+(defface jinx-highlight
+  '((t :inherit isearch))
+  "Face used to highlight current misspelling during correction.")
+
+(defface jinx-save
+  '((t :inherit font-lock-builtin-face))
+  "Face used for the save actions during correction.")
+
+(defface jinx-key
+  '((t :inherit completions-annotations))
+  "Face used for the select key during correction.")
+
+(defface jinx-annotation
+  '((t :inherit completions-annotations))
+  "Face used for the annotation during correction.")
+
 ;;;; Keymaps
 
 (defvar-keymap jinx-misspelled-map
@@ -195,7 +212,8 @@ checking."
   "SPC" #'self-insert-command
   "M-n" #'jinx-correct-next
   "M-p" #'jinx-correct-previous
-  "M-$" #'jinx-correct-next)
+  "M-$" #'jinx-correct-previous
+  "0 <t>" #'jinx-correct-select)
 (dotimes (i 9)
   (define-key jinx-correct-map (vector (+ ?1 i)) #'jinx-correct-select))
 
@@ -245,6 +263,16 @@ Predicate may return a position to skip forward.")
 (defvar-local jinx--session-words nil
   "List of words accepted in this session.")
 
+(defvar jinx--select-keys
+  "123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
+  "Quick select keys used by `jinx-correct'.")
+
+(defvar jinx--save-keys
+  `((?@ . ,#'jinx--save-personal)
+    (?* . ,#'jinx--save-file)
+    (?+ . ,#'jinx--save-session))
+  "Keys for save actions used by `jinx-correct'.")
+
 ;;;; Declarations for the bytecode compiler
 
 (defvar jinx-mode)
@@ -409,9 +437,7 @@ If VISIBLE is non-nil, only include visible overlays."
     (setq overlays
           (sort overlays
                 (lambda (a b) (< (overlay-start a) (overlay-start b)))))
-    (while (and (cdr overlays)
-                (> (abs (- (overlay-end (car overlays)) pt))
-                   (abs (- (overlay-start (cadr overlays)) pt))))
+    (while (and (cdr overlays) (<= (overlay-start (cadr overlays)) pt))
       (push (pop overlays) before))
     (nconc overlays (nreverse before))))
 
@@ -486,12 +512,15 @@ If VISIBLE is non-nil, only include visible overlays."
     (let* ((mod-name (file-name-with-extension "jinx-mod" module-file-suffix))
            (mod-file (locate-library mod-name t)))
       (unless mod-file
-        (let* ((c-name (file-name-with-extension mod-name ".c"))
+        (let* ((cc (or (getenv "CC")
+                       (seq-find #'executable-find '("gcc" "clang" "cc"))
+                       (error "Jinx: No C compiler found")))
+               (c-name (file-name-with-extension mod-name ".c"))
                (default-directory (file-name-directory
                                    (or (locate-library c-name t)
                                        (error "Jinx: %s not found" c-name))))
                (command
-                `("cc" "-I." "-O2" "-Wall" "-Wextra" "-fPIC" "-shared"
+                `(,cc "-I." "-O2" "-Wall" "-Wextra" "-fPIC" "-shared"
                   "-o" ,mod-name ,c-name
                   ,@(split-string-and-unquote
                      (condition-case nil
@@ -530,7 +559,7 @@ If VISIBLE is non-nil, only include visible overlays."
             (dolist (ov (overlays-in (pos-bol) (pos-eol)))
               (let ((inv (overlay-get ov 'invisible)))
                 (when (and (invisible-p inv) (overlay-get ov 'isearch-open-invisible))
-                  (push (if-let (fun (overlay-get ov 'isearch-open-invisible-temporary))
+                  (push (if-let ((fun (overlay-get ov 'isearch-open-invisible-temporary)))
                             (progn
                               (funcall fun ov nil)
                               (lambda () (funcall fun ov t)))
@@ -567,45 +596,45 @@ If VISIBLE is non-nil, only include visible overlays."
 
 (defun jinx--correct-suggestions (word)
   "Retrieve suggestions for WORD."
-  (delete-dups
-   (nconc
+  (nconc
+   (cl-loop
+    with idx = 1
+    with ht = (make-hash-table :test #'equal)
+    for dict in jinx--dicts
+    for desc = (jinx--mod-describe dict)
+    for group = (format "Suggestions from dictionary ‘%s’ (%s)" (car desc) (cdr desc))
+    nconc
     (cl-loop
-     with idx = 1
-     for dict in jinx--dicts nconc
-     (let* ((suggs (jinx--mod-suggest dict word))
-            (desc (jinx--mod-describe dict))
-            (group (format "Suggestions from dictionary ‘%s’ (%s)"
-                           (car desc) (cdr desc))))
-       (dolist (sugg suggs suggs)
-         (add-text-properties
-          0 (length sugg)
-          (list 'jinx--group group
-                'jinx--annotation (and (< idx 10) (format " (%s)" idx)))
-          sugg)
-         (cl-incf idx))))
-    (cl-loop
-     for dict in jinx--dicts for idx from 1 nconc
-     (let* ((at (propertize (make-string idx ?@)
-                            'face 'jinx-accept
-                            'rear-nonsticky t))
-            (desc (jinx--mod-describe dict))
-            (group "Accept and save word")
-            (ann (format " [Personal dictionary ‘%s’]" (car desc))))
-       (list (propertize (concat at word)
-                         'jinx--group group 'jinx--annotation ann)
-             (propertize (concat at (downcase word))
-                         'jinx--group group 'jinx--annotation ann))))
-    (list
-     (propertize (concat #("*" 0 1 (face jinx-accept rear-nonsticky t)) word)
-                 'jinx--group "Accept and save word"
-                 'jinx--annotation " [File]")
-     (propertize (concat #("+" 0 1 (face jinx-accept rear-nonsticky t)) word)
-                 'jinx--group "Accept and save word"
-                 'jinx--annotation " [Session]")))))
+     for sugg in (jinx--mod-suggest dict word)
+     if (not (gethash sugg ht)) collect
+     (progn
+       (add-text-properties
+        0 (length sugg)
+        (list 'jinx--group group
+              'jinx--prefix
+              (cond ((< idx 10)
+                     (format #("%d " 0 3 (face jinx-key)) idx))
+                    ((< (- idx 10) (length jinx--select-keys))
+                     (format #("0%c " 0 4 (face jinx-key)) (aref jinx--select-keys (- idx 10))))))
+        sugg)
+       (cl-incf idx)
+       (puthash sugg t ht)
+       sugg)))
+   (cl-loop for (key . fun) in jinx--save-keys nconc
+            (ensure-list (funcall fun nil key word)))))
 
-(defun jinx--correct-annotation (word)
-  "Annotate WORD during completion."
-  (get-text-property 0 'jinx--annotation word))
+(defun jinx--correct-affixation (cands)
+  "Affixate CANDS during completion."
+  (cl-loop for cand in cands collect
+           (list cand
+                 (or (get-text-property 0 'jinx--prefix cand) "")
+                 (or (get-text-property 0 'jinx--suffix cand) ""))))
+
+(defun jinx--correct-annotation (cand)
+  "Annotate CAND during completion."
+  (if-let ((prefix (get-text-property 0 'jinx--prefix cand)))
+      (format #(" (%s)" 0 5 (face jinx-key)) (string-trim prefix))
+    (get-text-property 0 'jinx--suffix cand)))
 
 (defun jinx--correct-group (word transform)
   "Group WORD during completion, TRANSFORM candidate if non-nil."
@@ -622,6 +651,7 @@ If VISIBLE is non-nil, only include visible overlays."
                    (cycle-sort-function . identity)
                    (category . jinx)
                    (group-function . jinx--correct-group)
+                   (affixation-function . jinx--correct-affixation)
                    (annotation-function . jinx--correct-annotation))
       (complete-with-action action word str pred))))
 
@@ -629,7 +659,7 @@ If VISIBLE is non-nil, only include visible overlays."
   "Correct word at OVERLAY, maybe RECENTER and show prompt INFO."
   (let* ((word (buffer-substring-no-properties
                 (overlay-start overlay) (overlay-end overlay)))
-         (selected
+         (choice
           (jinx--correct-highlight overlay
             (lambda ()
               (when recenter (recenter))
@@ -639,34 +669,19 @@ If VISIBLE is non-nil, only include visible overlays."
                      (format "Correct ‘%s’%s: " word (or info ""))
                      (jinx--correct-table word)
                      nil nil nil t word)
-                    word))))))
-    (if (string-match-p "\\`[@+*]" selected)
-        (let* ((new-word (replace-regexp-in-string "\\`[@+*]+" "" selected))
-               (idx (- (length selected) (length new-word) 1)))
-          (when (equal new-word "") (setq new-word word))
-          (cond
-           ((string-prefix-p "+" selected)
-            (add-to-list 'jinx--session-words new-word))
-           ((string-prefix-p "*" selected)
-            (add-to-list 'jinx--session-words new-word)
-            (setq jinx-local-words
-                  (string-join
-                   (sort (delete-dups
-                          (cons new-word (split-string jinx-local-words)))
-                         #'string<)
-                   " "))
-            (add-file-local-variable 'jinx-local-words jinx-local-words))
-           (t (jinx--mod-add (or (nth idx jinx--dicts)
-                                 (user-error "Invalid dictionary"))
-                             new-word)))
+                    word)))))
+         (len (length choice)))
+    (if-let ((save (and (> len 0) (assq (aref choice 0) jinx--save-keys))))
+        (progn
+          (funcall (cdr save) 'save (car save) (if (> len 1) (substring choice 1) word))
           (jinx--recheck-overlays))
-      (when-let (((not (equal selected word)))
+      (when-let (((not (equal choice word)))
                  (start (overlay-start overlay))
                  (end (overlay-end overlay)))
         (undo-boundary)
         (delete-overlay overlay)
         (goto-char end)
-        (insert-before-markers selected)
+        (insert-before-markers choice)
         (delete-region start end)))))
 
 (defun jinx--load-dicts ()
@@ -684,33 +699,86 @@ If VISIBLE is non-nil, only include visible overlays."
   (modify-syntax-entry ?' "w" jinx--syntax-table)
   (modify-syntax-entry ?. "." jinx--syntax-table))
 
+;;;; Save functions
+
+(defun jinx--save-action (key word ann)
+  "Format save action given KEY, WORD and ANN."
+  (propertize
+   (concat (propertize (if (stringp key) key (char-to-string key))
+                       'face 'jinx-save 'rear-nonsticky t)
+           word)
+   'jinx--group "Accept and save word"
+   'jinx--suffix (format #(" [%s]" 0 5 (face jinx-annotation)) ann)))
+
+(defun jinx--save-personal (save key word)
+  "Save WORD in personal dictionary.
+If SAVE is non-nil save, otherwise format candidate given action KEY."
+  (if save
+      (let ((idx (seq-position word key (lambda (x y) (not (equal x y))))))
+        (jinx--mod-add (or (nth idx jinx--dicts)
+                           (user-error "Invalid dictionary"))
+                       (substring word idx)))
+    (cl-loop
+     for dict in jinx--dicts for idx from 1
+     for at = (make-string idx key)
+     for ann = (format "Personal ‘%s’" (car (jinx--mod-describe dict))) nconc
+     (delete-consecutive-dups
+      (list (jinx--save-action at word ann)
+            (jinx--save-action at (downcase word) ann))))))
+
+(defun jinx--save-file (save key word)
+  "Save WORD in file-local variable.
+If SAVE is non-nil save, otherwise format candidate given action KEY."
+  (if save
+      (progn
+        (add-to-list 'jinx--session-words word)
+        (setq jinx-local-words
+              (string-join
+               (sort (delete-dups
+                      (cons word (split-string jinx-local-words)))
+                     #'string<)
+               " "))
+        (add-file-local-variable 'jinx-local-words jinx-local-words))
+    (jinx--save-action key word "File")))
+
+(defun jinx--save-session (save key word)
+  "Save WORD for the current session.
+If SAVE is non-nil save, otherwise format candidate given action KEY."
+  (if save
+      (add-to-list 'jinx--session-words word)
+    (jinx--save-action key word "Session")))
+
 ;;;; Public commands
 
 ;;;###autoload
-(defun jinx-languages (&optional global)
-  "Change languages locally or globally.
-With prefix argument GLOBAL non-nil change the languages globally."
-  (interactive "*P")
-  (jinx--load-module)
-  (when-let ((langs
-              (completing-read-multiple
-               (format "Change languages (%s): "
-                       (string-join (split-string jinx-languages) ", "))
-               (delete-dups (jinx--mod-langs)) nil t)))
-    (setq langs (string-join langs " "))
-    (cond
-     (global
-      (kill-local-variable 'jinx-languages)
-      (setq-default jinx-languages langs))
-     (t
-      (setq-local jinx-languages langs)
-      (when (or (assq 'jinx-languages file-local-variables-alist)
-                (and buffer-file-name
-                     (y-or-n-p "Save `jinx-languages' as file-local variable? ")))
-        (add-file-local-variable 'jinx-languages jinx-languages)
-        (setf (alist-get 'jinx-languages file-local-variables-alist) jinx-languages))))
-    (jinx--load-dicts)
-    (jinx--cleanup)))
+(defun jinx-languages (langs &optional global)
+  "Set languages locally or globally to LANGS.
+With prefix argument GLOBAL change the languages globally."
+  (interactive
+   (list
+    (progn
+      (jinx--load-module)
+      (string-join
+       (or (completing-read-multiple
+            (format "Change languages (%s): "
+                    (string-join (split-string jinx-languages) ", "))
+            (delete-dups (jinx--mod-langs)) nil t)
+           (user-error "No languages selected"))
+       " "))
+    current-prefix-arg))
+  (cond
+   (global
+    (kill-local-variable 'jinx-languages)
+    (setq-default jinx-languages langs))
+   (t
+    (setq-local jinx-languages langs)
+    (when (or (assq 'jinx-languages file-local-variables-alist)
+              (and buffer-file-name
+                   (y-or-n-p "Save `jinx-languages' as file-local variable? ")))
+      (add-file-local-variable 'jinx-languages jinx-languages)
+      (setf (alist-get 'jinx-languages file-local-variables-alist) jinx-languages))))
+  (jinx--load-dicts)
+  (jinx--cleanup))
 
 ;;;###autoload
 (defun jinx-correct (&optional all)
@@ -753,12 +821,16 @@ If prefix argument ALL non-nil correct all misspellings."
 (defun jinx-correct-select ()
   "Quick selection key for corrections."
   (interactive)
-  (let ((word (nth (- last-input-event ?1)
-                   (all-completions "" minibuffer-completion-table))))
-    (when (and word (not (string-match-p "\\`[@+*]" word)))
-      (delete-minibuffer-contents)
-      (insert word)
-      (exit-minibuffer))))
+  (let* ((keys (this-command-keys-vector))
+         (word (nth (if (eq (aref keys 0) ?0)
+                        (+ 9 (or (seq-position jinx--select-keys (aref keys 1)) 999))
+                      (- (aref keys 0) ?1))
+                    (all-completions "" minibuffer-completion-table))))
+    (unless (and word (get-text-property 0 'jinx--prefix word))
+      (user-error "Invalid select key `%s'" (key-description keys)))
+    (delete-minibuffer-contents)
+    (insert word)
+    (exit-minibuffer)))
 
 (defun jinx-correct-next (n)
   "Skip to Nth next misspelling."
@@ -789,8 +861,7 @@ If prefix argument ALL non-nil correct all misspellings."
           jinx--include-faces (jinx--mode-list jinx-include-faces)
           jinx--exclude-faces (jinx--mode-list jinx-exclude-faces)
           jinx--camel (or (eq jinx-camel-modes t)
-                          (cl-loop for m in jinx-camel-modes
-                                   thereis (derived-mode-p m)))
+                          (seq-some #'derived-mode-p jinx-camel-modes))
           jinx--session-words (split-string jinx-local-words))
     (jinx--load-dicts)
     (add-hook 'window-state-change-hook #'jinx--reschedule nil t)
